package com.emato.eoss.api;


import com.emato.eoss.manager.oss.minio.utils.MinioUtils;
import com.emato.eoss.msg.resp.Result;
import com.google.common.collect.ImmutableBiMap;
import io.minio.ObjectWriteResponse;
import io.minio.StatObjectResponse;
import io.minio.http.Method;
import okhttp3.Headers;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * 文件存储API
 *
 * @author Scott Chen
 * @date 2021-03-09
 */
@RestController
public class OssApi {
    private static final Logger logger = LoggerFactory.getLogger(OssApi.class);

    @Autowired
    private MinioUtils minioUtils;

    @Value("${minio.buckets.product}")
    private String bucketName;

    /**
     * 上传文件
     * MultipartFile 方式上传
     *
     * @param file           上传文件
     * @param bucket         OSS存储桶
     * @param targetFileName 另存为的文件名，如果为 USE_ORIGIN，则使用原文件名存储
     * @param request
     * @param response
     * @return
     */
    @RequestMapping("/upload")
    public Result upload(@RequestParam(value = "file") MultipartFile file,
                         @RequestParam(value = "bucket") String bucket,
                         @RequestParam(value = "targetFileName") String targetFileName,
                         HttpServletRequest request, HttpServletResponse response) {

        bucket = StringUtils.isBlank(bucket) ? bucketName : bucket;

        String objectName = null;
        objectName = file.getOriginalFilename();

        if (StringUtils.isNoneBlank(targetFileName) && !targetFileName.equals("USE_ORIGIN")) {
            // 使用另存为文件名进行存储
            objectName = targetFileName;
        } else {
            if (StringUtils.isBlank(objectName)) {
                return Result.error("上传文件名为空");
            }
        }

        String contentType = file.getContentType();

        try {
            InputStream is = file.getInputStream();
            ObjectWriteResponse owr = minioUtils.putFileWithInputStream(bucket, objectName, is,
                    null, null, contentType);

            String object = owr.object();
            logger.info("{}上传到{}成功。", object, bucket);

            Map<String, String> map = new HashMap<>();
            map.put("bucket", bucket);
            map.put("fileName", object);
            return Result.success(map);
        } catch (Exception e) {
            String info = "上传文件" + objectName + "出错";
            logger.error(info);
            return Result.error(info);
        }
    }

    /**
     * 上传文件
     * 流方式上传
     *
     * @param bucket OSS存储桶
     * @param objectName OSS存储对象
     * @param byteFile 上传文件字节码
     * @param request
     * @param response
     * @return
     */
    @RequestMapping("/uploadWithByte")
    public Result upload(@RequestParam(value="bucket") String bucket,
                         @RequestParam(value="objectName") String objectName,
                         @RequestParam(value="inputStream") byte[] byteFile,
                         HttpServletRequest request, HttpServletResponse response) {
        try{
            bucket = StringUtils.isBlank(bucket) ? bucketName : bucket;

            if (byteFile == null) {
                return Result.error("上传文件byteFile为空");
            }
            ByteArrayInputStream inputStream = new ByteArrayInputStream(byteFile);

            ObjectWriteResponse owr = minioUtils.putFileWithInputStream(bucket, objectName, inputStream,
                    null, null, null);

            String region = owr.region();
            String object = owr.object();
            Headers headers = owr.headers();

            logger.info("{}上传到{}成功。", object, bucket);

            Map<String, String> map = new HashMap<>();
            map.put("bucket", bucket);
            map.put("fileName", object);
            return Result.success(map);
        } catch (Exception e) {
            String info = "上传文件" + objectName + "出错";
            logger.error(info);
            return Result.error(info);
        }
    }

    /**
     * 上传文件
     * 流方式上传
     *
     * @param bucket OSS存储桶
     * @param objectName OSS存储对象
     * @param inputStream 上传文件流
     * @param request
     * @param response
     * @return
     */
    @RequestMapping("/uploadWithInputStream")
    public Result upload(@RequestParam(value="bucket") String bucket,
                        @RequestParam(value="objectName") String objectName,
                        @RequestParam(value="inputStream") InputStream inputStream,
                         HttpServletRequest request, HttpServletResponse response) {
        try{
            bucket = StringUtils.isBlank(bucket) ? bucketName : bucket;

            ObjectWriteResponse owr = minioUtils.putFileWithInputStream(bucket, objectName, inputStream,
                    null, null, null);

            String region = owr.region();
            String object = owr.object();
            Headers headers = owr.headers();

            logger.info("{}上传到{}成功。", object, bucket);

            Map<String, String> map = new HashMap<>();
            map.put("bucket", bucket);
            map.put("fileName", object);
            return Result.success(map);
        } catch (Exception e) {
            String info = "上传文件" + objectName + "出错";
            logger.error(info);
            return Result.error(info);
        }
    }

    /**
     * 下载文件
     *
     * @param bucket         OSS存储桶
     * @param object         OSS存储对象
     * @param outputFileName 另存为文件名
     * @param request
     * @param response
     * @return
     */
    @RequestMapping("/download")
    public void download(@RequestParam(value = "bucket") String bucket,
                         @RequestParam(value = "object") String object,
                         @RequestParam(value = "outputFileName", required = false) String outputFileName,
                         HttpServletRequest request, HttpServletResponse response) {

        if (StringUtils.isBlank(outputFileName)) {
            outputFileName = object;
        }

        OutputStream os = null;
        try {
            StatObjectResponse stat = minioUtils.getObjectInfo(bucket, object);
            byte[] buffer = minioUtils.getObjectWithByte(bucket, object);

            response.setHeader("Content-Disposition", "attachment;filename=" + outputFileName);
            response.setCharacterEncoding("UTF-8");
            response.setContentLength((int) stat.size());
            response.setContentType(MediaType.APPLICATION_OCTET_STREAM_VALUE);

            os = response.getOutputStream();
            os.write(buffer);
            os.close();
            logger.info("从{}下载{}文件成功，另存为: {}，长度: {}, 文件类型: {}", bucket, object, outputFileName, (int)stat.size(), stat.contentType());
        } catch (Exception e) {
            String info = "下载文件" + object + "出错";
            logger.error(info, e);
        } finally {
            if (os != null) {
                try {
                    os.close();
                } catch (IOException e) {
                    logger.error("下载文件" + object + "，关闭OutputStream异常", e);
                }
            }
        }
    }

    /**
     * 下载到服务端指定路径
     * 服务端指定路径为OSS系统部署的服务端
     *
     * @param bucket OSS存储桶
     * @param object OSS存储对象
     * @param targetPath 另存为路径
     * @param targetFileName 另存为文件名
     * @param request
     * @param response
     * @return
     */
    @RequestMapping("/downloadObject")
    public Result downloadObject(@RequestParam(value="bucket") String bucket,
                                 @RequestParam(value="object") String object,
                                 @RequestParam(value="targetPath", required=false) String targetPath,
                                 @RequestParam(value="targetFileName", required=false) String targetFileName,
                                 HttpServletRequest request, HttpServletResponse response) {
        if (StringUtils.isBlank(bucket)) {
            return Result.error("bucket为空");
        }
        if (StringUtils.isBlank(object)) {
            return Result.error("object为空");
        }

        File file = new File(targetPath);
        if (!file.exists()) {
            file.mkdirs();
        }
        // 输出文件全路径
        String outputFullPath = targetPath + File.pathSeparator + targetFileName;

        try{
            minioUtils.downloadObject(
                    bucket, object, outputFullPath);

            logger.info("从{}下载{}文件成功，已另存为{}", bucket, object, outputFullPath);
            return Result.success("从" + bucket + "下载" + object + "文件成功，已另存为" + outputFullPath);
        } catch (Exception e) {
            String info = "下载文件" + object + "并另存为" + outputFullPath + "出错";
            logger.error(info);
            return Result.error(info);
        }
    }

    /**
     * 获取对象Url
     *
     * expires为0时，表示永久Url
     *
     * @param bucket OSS存储桶
     * @param object OSS存储对象
     * @param method Url访问方法
     * @param expires Url有效期，默认为天，如果为0，表示永久有效
     * @param extraQueryParams Url的请求参数
     *                         如：response-content-type=application/json
     *                              response-content-type 为返回数据类型key
     *                              application/json 为返回数据类型value
     * @return
     */
    @RequestMapping("/objectUrl")
    public Result getObjectUrl(@RequestParam(value = "bucket") String bucket,
                               @RequestParam(value = "object") String object,
                               @RequestParam(value = "method", required = false) String method,
                               @RequestParam(value = "expires") Integer expires,
                               @RequestBody(required = false) Map<String, String> extraQueryParams) {

        Method reqMethod = null;
        if (StringUtils.isNoneBlank(method)) {
            for (Method met : Method.values()) {
                if (method.equals(met.name())) {
                    reqMethod = met;
                }
            }
        }
        reqMethod = reqMethod == null ? Method.GET : reqMethod;

        if (expires <= -1) {
            String info = "请求" + bucket + "内对象" + object + "Url时，有效期不能为负数";
            logger.error(info);
            return Result.error(info);
        }

        String url = null;

        if (expires == 0) {
            // 请求永久Url
            try {
                url = minioUtils.getPersistentObjectUrl(bucket, object, reqMethod, extraQueryParams);
            } catch (Exception e) {
                String info = "请求" + bucket + "内对像" + object + "永久Url出错";
                logger.error(info, e);
                return Result.error(info);
            }
        } else {
            try {
                url = minioUtils.getObjectUrl(bucket, object, reqMethod, expires, TimeUnit.DAYS, extraQueryParams);
            } catch (Exception e) {
                String info = "请求" + bucket + "内对像" + object + "Url出错";
                logger.error(info, e);
                return Result.error(info);
            }
        }

        ImmutableBiMap data = ImmutableBiMap.of("bucket", bucket, "object", object, "url", url);
        return Result.success(data);
    }


}
