package com.emato.eoss.manager.oss.minio.utils;

import com.emato.eoss.manager.oss.minio.config.MinioProp;
import io.minio.*;
import io.minio.errors.MinioException;
import io.minio.http.Method;
import io.minio.messages.Bucket;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;

import javax.annotation.Nonnull;
import javax.crypto.SecretKey;
import java.io.IOException;
import java.io.InputStream;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * minio 工具类
 *
 * @author Scott Chen
 * @date 2021-03-04
 */
public class MinioUtils {
    private static final Logger logger = LoggerFactory.getLogger(MinioUtils.class);

    private static final ThreadLocal<MinioClient> minioClientThreadLocal = new ThreadLocal<>();
    private static final ThreadLocal<MinioUtils> minioUtilsThreadLocal = new ThreadLocal<>();

    private MinioProp minioProp = null;


    private MinioUtils() {}

    public static MinioUtils create(@Nonnull MinioProp minioProp){
        MinioUtils minioUtils = minioUtilsThreadLocal.get();
        if (minioUtils == null) {
            synchronized (MinioUtils.class) {
                if (minioUtils == null) {
                    minioUtils = MinioUtils.builder().minioProp(minioProp).build();
                    minioUtilsThreadLocal.set(minioUtils);
                }
            }
        }
        return minioUtils;
    }

    public MinioClient minioClient() {
        MinioClient client = minioClientThreadLocal.get();
        if (client == null) {
            synchronized (MinioUtils.class) {
                if (client == null) {
                    client = MinioClient.builder()
                            .endpoint(minioProp.getEndpoint(),
                                    minioProp.getPort(),
                                    minioProp.getSecure())
                            .credentials(minioProp.getAccesskeyId(),
                                    minioProp.getSecretAccessKey())
                            .build();
                    minioClientThreadLocal.set(client);
                }
            }
        }
        return client;
    }

    
    // ------------------------------ Bucket ------------------------------

    /**
     * 存储桶是否存在
     *
     * @param bucket
     * @return
     * @throws MinioException
     */
    public boolean existBucket(String bucket) throws MinioException {
        boolean found = true;
        try {
            found = minioClient().bucketExists(
                    BucketExistsArgs.builder()
                            .bucket(bucket)
                            .build());
        } catch (Exception e) {
            logger.error("judge bucket error.", e);
            throw new MinioException(e.getMessage());
        }
        return found;
    }

    /**
     * 创建存储桶
     *
     * @param bucket 存储桶名 @NotNull
     * @return
     * @date 2021-03-05
     **/
    public void createBucketByNotExists(String bucket) throws MinioException {
        if (!existBucket(bucket)) {
            try {
                minioClient().makeBucket(
                        MakeBucketArgs.builder()
                                .bucket(bucket)
                                .build());
            } catch (Exception e) {
                logger.error("create bucket error.", e);
                throw new MinioException(e.getMessage());
            }
        }
    }

    /**
     * 获取bucket
     *
     * @param bucket
     * @return
     * @throws MinioException
     */
    public Optional<Bucket> getBucket(String bucket) throws MinioException {
        try {
            List<Bucket> bucketList = getAllBuckets(Collections.emptyMap());
            return bucketList.stream().filter(i -> i.name().equals(bucket)).findFirst();
        } catch (Exception e) {
            logger.error("get bucket error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 获取全部bucket
     *
     * @param headers
     * @return
     * @throws MinioException
     */
    public List<Bucket> getAllBuckets(Map<String, String> headers) throws MinioException {
        try {
            return minioClient().listBuckets(
                    ListBucketsArgs.builder()
                            .extraHeaders(headers)
                            .build());
        } catch (Exception e) {
            logger.error("get all bucket error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 删除bucketName
     *
     * @param bucket bucket名称
     */
    public void removeBucket(String bucket) throws MinioException {
        try {
            minioClient().removeBucket(
                    RemoveBucketArgs.builder()
                            .bucket(bucket)
                            .build());
        } catch (Exception e) {
            logger.error("remove bucket error.", e);
            throw new MinioException(e.getMessage());
        }
    }



    // ------------------------------ Object ------------------------------

    /**
     * 获取对象数据
     * 在释放网络资源时，必须释放InputStream流
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @return java.io.InputStream
     * @date 2021-03-05
     **/
    public InputStream getObject(String bucket, String objectName) throws MinioException{
        try {
            return getObject(bucket, objectName, null, null, null);
        } catch (Exception e) {
            logger.error("get object error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 获取对象数据
     * 在释放网络资源时，必须释放InputStream流
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @param offset 获取开始位置
     * @param length 获取长度
     * @param secretKey secretKey
     * @return
     * @throws MinioException
     */
    public InputStream getObject(String bucket, String objectName,
                                 Long offset, Long length,
                                 SecretKey secretKey) throws MinioException{
        try{
            GetObjectArgs.Builder builder = GetObjectArgs.builder();
            builder.bucket(bucket);
            builder.object(objectName);
            if (offset != null && offset >= 0) {
                builder.offset(offset);
            }
            if (length != null && length > 0) {
                builder.length(length);
            }
            if (secretKey != null) {
                builder.ssec(new ServerSideEncryptionCustomerKey(secretKey));
            }
            return minioClient().getObject(builder.build());
        } catch (Exception e) {
            logger.error("get object with length and ssec error.", e);
            throw new MinioException(e.getMessage());
        }
    }



    /**
     * 字节方式获取对象数据
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @return byte[]
     * @date 2021-03-05
     **/
    public byte[] getObjectWithByte(String bucket, String objectName) throws MinioException{
        try {
            return getObjectWithByte(bucket, objectName, null, null, null);
        } catch (Exception e) {
            logger.error("get file object return byte error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 字节方式获取对象数据
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @param offset 获取开始位置
     * @param length 获取长度
     * @param secretKey secretKey
     * @return byte[]
     * @date 2021-03-05
     **/
    public byte[] getObjectWithByte(String bucket, String objectName,
                                         Long offset, Long length,
                                         SecretKey secretKey) throws MinioException{
        ArrayList<byte[]> partList = new ArrayList<>();
        InputStream is = null;
        try {
            is = getObject(bucket, objectName, offset, length, secretKey);

            byte[] buffered = new byte[1024];
            int len;
            while ((len = is.read(buffered)) > 0) {
                byte[] temp = new byte[len];
                System.arraycopy(buffered, 0, temp, 0, len);
                partList.add(temp);
            }
            is.close();
        } catch (Exception e) {
            logger.error("get file object with length ssec return error.", e);
            throw new MinioException(e.getMessage());
        } finally {
            if (is != null) {
                try {
                    is.close();
                } catch (IOException e) {
                    logger.error("下载文件" + objectName + "，关闭InputStream异常", e);
                }
            }
        }

        if (partList.isEmpty()) {
            return new byte[]{0};
        }

        Integer size = partList.stream().map(x -> x.length).reduce(0, Integer::sum);
        byte[] result = new byte[size];
        int currentIndex = 0;
        for (byte[] bytes : partList) {
            for (byte aByte : bytes) {
                result[currentIndex++] = aByte;
            }
        }
        return result;
    }

    /**
     * 将对象的数据下载到文件
     * 下载到服务端指定路径
     * objectName下载为outputFileName
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @param outputFileName 输出文件，可以包含路径 @NotNull
     * @return
     * @date 2021-03-05
     **/
    public void downloadObject(String bucket, String objectName, String outputFileName) throws MinioException{
        try {
            DownloadObjectArgs.Builder builder = DownloadObjectArgs.builder();
            builder.bucket(bucket);
            builder.object(objectName);

            outputFileName = StringUtils.isBlank(outputFileName) ? objectName : outputFileName;
            builder.filename(outputFileName);

            minioClient().downloadObject(builder.build());
        } catch (Exception e) {
            logger.error("download object error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 将对象的数据下载到文件
     * 下载到服务端指定路径
     * objectName下载为outputFileName
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @param outputFileName 输出文件，可以包含路径 @NotNull
     * @param secretKey secretKey @NotNull
     * @return
     * @date 2021-03-05
     **/
    public void downloadObjectBySsec(String bucket, String objectName,
                                     String outputFileName, SecretKey secretKey) throws MinioException{
        try {
            if (secretKey == null) {
                String info = "download object error. secretKey is not null.";
                logger.error(info);
                throw new MinioException(info);
            }
            DownloadObjectArgs.Builder builder = DownloadObjectArgs.builder();
            builder.bucket(bucket);
            builder.object(objectName);
            builder.ssec(new ServerSideEncryptionCustomerKey(secretKey));
            builder.filename(outputFileName);

            minioClient().downloadObject(builder.build());
        } catch (Exception e) {
            logger.error("download object with ssec error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 流式上传文件
     * 流putFileWithInputStream存储为objectName
     * 存储桶不存在会抛异常
     *
     * @param bucket 桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull 多路径存储实例： package/subpackage/.../objectName 开头不要有/
     * @param putFileWithInputStream 上传文件输入流 @NotNull
     * @param objectSize 文件大小
     * @param partSize 分片大小
     * @param contentType 文本类型 默认类型：application/octet-stream
     * @return
     * @date 2021-03-05
     **/
    public ObjectWriteResponse putFileWithInputStreamExistBucket(String bucket, String objectName, InputStream putFileWithInputStream,
                                                                 Long objectSize, Long partSize, String contentType) throws MinioException {
        if (!existBucket(bucket)) {
            throw new MinioException("The bucket not exists.");
        }
        return putFileWithInputStream(bucket, objectName, putFileWithInputStream, objectSize, partSize, contentType);
    }

    /**
     * 流式上传文件
     * 流putFileWithInputStream存储为objectName
     * 存储桶不存在会创建
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull 多路径存储实例： package/subpackage/.../objectName 开头不要有/
     * @param putFileWithInputStream 上传文件输入流 @NotNull
     * @param objectSize 文件大小
     * @param partSize 分片大小
     * @param contentType 文本类型 默认类型：application/octet-stream
     * @return
     * @date 2021-03-05
     **/
    public ObjectWriteResponse putFileWithInputStream(String bucket, String objectName, InputStream putFileWithInputStream,
                           Long objectSize, Long partSize, String contentType) throws MinioException {
        createBucketByNotExists(bucket);
        try {
            objectSize = objectSize == null ? 0 : objectSize;
            partSize = partSize == null ? 0 : partSize;
            if (objectSize <= 0 && partSize <= 0) {
                objectSize = Long.valueOf(putFileWithInputStream.available());
                partSize = -1L;
            }

            if (StringUtils.isBlank(contentType)) {
                contentType = MediaType.APPLICATION_OCTET_STREAM_VALUE;
            }

            PutObjectArgs.Builder builder = PutObjectArgs.builder();
            builder.bucket(bucket);
            builder.object(objectName);
            builder.stream(putFileWithInputStream, objectSize, partSize);
            builder.contentType(contentType);
            return minioClient().putObject(builder.build());
        } catch (Exception e) {
            logger.error("put file error.", e);
            throw new MinioException(e.getMessage());
        }
    }


    /**
     * 文件地址上传文件
     * 存储桶不存在抛出异常
     *
     * @param bucket 存储桶名称 @NotNull
     * @param uploadFileName 上传文件 @NotNull
     * @param contentType 文本类型 默认类型：application/octet-stream
     * @return
     * @date 2021-03-05
     **/
    public ObjectWriteResponse uploadFile(String bucket, String uploadFileName,
                                                  String contentType) throws MinioException {
        if (!existBucket(bucket)) {
            throw new MinioException("The bucket not exists.");
        }
        if (!existBucket(uploadFileName)) {
            throw new MinioException("The upload file name is null.");
        }
        // uploadFileName存储为objectName
        String objectName = uploadFileName;
        return uploadFileSaveAsObject(bucket, objectName, uploadFileName, contentType);
    }

    /**
     * 文件上传文件
     * uploadFileName存储为objectName
     * 存储桶不存在抛出异常
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull 多路径存储实例： package/subpackage/.../objectName 开头不要有/
     * @param uploadFileName 上传文件 @NotNull
     * @param contentType 文本类型 默认类型：application/octet-stream
     * @return
     * @date 2021-03-05
     **/
    public ObjectWriteResponse uploadFileSaveAsObject(String bucket, String objectName,
                                                      String uploadFileName, String contentType) throws MinioException {
        if (!existBucket(bucket)) {
            throw new MinioException("The bucket not exists.");
        }
        if (!existBucket(uploadFileName)) {
            throw new MinioException("The upload file name is null.");
        }

        try {
            UploadObjectArgs.Builder builder = UploadObjectArgs.builder();
            builder.bucket(bucket);
            builder.filename(uploadFileName);

            objectName = StringUtils.isBlank(objectName) ? uploadFileName : objectName;
            builder.object(objectName);

            contentType = StringUtils.isNotBlank(contentType) ? contentType : MediaType.APPLICATION_OCTET_STREAM_VALUE;
            builder.contentType(contentType);

            return minioClient().uploadObject(builder.build());
        } catch (Exception e) {
            logger.error("upload file error.", e);
            throw new MinioException(e.getMessage());
        }
    }

    /**
     * 删除文件
     *
     * @param bucket 存储桶名称 @NotNull
     * @param objectName 存储对象名称，在存储桶内存储的对象名称 @NotNull
     * @return
     * @operation add
     * @date 2020/10/30
     **/
    public void removeObject(String bucket, String objectName) throws MinioException {
        try {
            minioClient().removeObject(
                    RemoveObjectArgs.builder()
                            .bucket(bucket)
                            .object(objectName)
                            .build());
        } catch (Exception e) {
            logger.error("remove file error.", e);
            throw new MinioException(e.getMessage());
        }
    }


    /**
     * 复制文件
     *
     * @param sourceBucket 源文件桶 @NotNull
     * @param sourceObject 源文件名 @NotNull
     * @param bucket 目标桶 @NotNull
     * @param object 目标文件 @NotNull
     * @return
     * @date 2020/10/30
     **/
    public void copyObject(String sourceBucket, String sourceObject,
                           String bucket, String object) throws MinioException{
        try {
            CopySource source = CopySource.builder().bucket(sourceBucket).object(sourceObject).build();
            minioClient().copyObject(
                    CopyObjectArgs.builder()
                            .bucket(bucket)
                            .object(object)
                            .source(source)
                            .build());
        } catch (Exception e) {
            logger.error("copy file error.", e);
            throw new MinioException(e.getMessage());
        }
    }


    /**
     * 获取存储对象指定有效期的Url
     *
     * url可以指定适不同HTTP Method用于不同发送请求
     * url可以指定有效期
     * 即使bucketName是私有的，该url也可以访问
     * {@link  Method}
     *
     * @param bucket bucket名称
     * @param objectName 文件名称
     * @param method HTTP method {@link Method}，默认为GET
     * @param expires 过期时间，默认为秒
     * @param timeUnit 过期时间的单位
     * @param extraQueryParams Url的请求参数
     *                         如：response-content-type=application/json
     *                              response-content-type 为返回数据类型key
     *                              application/json 为返回数据类型value
     * @throws Exception
     */
    public String getObjectUrl(String bucket, String objectName,
                               Method method, Integer expires, TimeUnit timeUnit,
                               Map<String, String> extraQueryParams) throws Exception {


        GetPresignedObjectUrlArgs.Builder builder = GetPresignedObjectUrlArgs.builder();

        method = method == null ? Method.GET : method;
        builder.method(method);
        builder.bucket(bucket);
        builder.object(objectName);

        expires = expires == null ? 7 : expires;
        timeUnit = timeUnit == null ? TimeUnit.DAYS : timeUnit;
        builder.expiry(expires, timeUnit);

        if (extraQueryParams != null && !extraQueryParams.isEmpty()) {
            builder.extraQueryParams(extraQueryParams);
        }
        return minioClient().getPresignedObjectUrl(builder.build());
    }

    /**
     * 获取存储对象永久的Url
     *
     * @param bucket           bucket名称
     * @param objectName       文件名称
     * @param method           HTTP method {@link Method}，默认为GET
     * @param extraQueryParams Url的请求参数
     *                         如：response-content-type=application/json
     *                         response-content-type 为返回数据类型key
     *                         application/json 为返回数据类型value
     * @return
     * @throws Exception
     */
    public String getPersistentObjectUrl(String bucket, String objectName,
                                         Method method, Map<String, String> extraQueryParams) throws Exception {


        GetPresignedObjectUrlArgs.Builder builder = GetPresignedObjectUrlArgs.builder();

        method = method == null ? Method.GET : method;
        builder.method(method);
        builder.bucket(bucket);
        builder.object(objectName);

        if (extraQueryParams != null && !extraQueryParams.isEmpty()) {
            builder.extraQueryParams(extraQueryParams);
        }
        // 去除文件Url地址中的全部参数，只保留问号(?)前面部分
        String temp_url = minioClient().getPresignedObjectUrl(builder.build());
        return temp_url.substring(0, temp_url.indexOf('?'));
    }

    /**
     * 获取存储对象信息
     *
     * @param bucket     bucket名称
     * @param objectName 文件名称
     * @throws Exception
     */
    public StatObjectResponse getObjectInfo(String bucket, String objectName) throws Exception {
        return minioClient().statObject(
                StatObjectArgs.builder()
                        .bucket(bucket)
                        .object(objectName)
                        .build());
    }

    /**
     * 获取存储对象信息
     * Get information of SSE-C encrypted object.
     *
     * @param bucket bucket名称
     * @param objectName 存储对象名称，在存储桶内存储的对象名称
     * @param secretKey secretKey
     * @throws Exception
     */
    public StatObjectResponse getObjectInfoWithSsec(String bucket, String objectName,
                                                SecretKey secretKey) throws Exception {
        if (secretKey == null) {
            String info = "get object error. secretKey is not null.";
            logger.error(info);
            throw new MinioException(info);
        }
        return minioClient().statObject(
                StatObjectArgs.builder()
                        .bucket(bucket)
                        .object(objectName)
                        .ssec(new ServerSideEncryptionCustomerKey(secretKey))
                        .build());
    }

    /**
     * 获取存储对象信息
     * Get information of a versioned object.
     *
     * @param bucket bucket名称
     * @param objectName 存储对象名称，在存储桶内存储的对象名称
     * @param versionId version id
     * @throws Exception
     */
    public StatObjectResponse getObjectInfoWithVersion(String bucket, String objectName,
                                                   String versionId) throws Exception {
        if (versionId == null) {
            String info = "get object info. versionId is not null.";
            logger.error(info);
            throw new MinioException(info);
        }
        return minioClient().statObject(
                StatObjectArgs.builder()
                        .bucket(bucket)
                        .object(objectName)
                        .versionId(versionId)
                        .build());
    }

    /**
     * 获取存储对象信息
     * Get information of a versioned object.
     *
     * @param bucket bucket名称
     * @param objectName 存储对象名称，在存储桶内存储的对象名称
     * @param versionId version id
     * @param secretKey secretKey
     * @throws Exception
     */
    public StatObjectResponse getObjectInfoWithVersionSsec(String bucket, String objectName,
                                                   String versionId, SecretKey secretKey) throws Exception {
        if (versionId == null) {
            String info = "get object info. versionId is not null.";
            logger.error(info);
            throw new MinioException(info);
        }
        if (secretKey == null) {
            String info = "get object info. secretKey is not null.";
            logger.error(info);
            throw new MinioException(info);
        }
        return minioClient().statObject(
                StatObjectArgs.builder()
                        .bucket(bucket)
                        .object(objectName)
                        .versionId(versionId)
                        .ssec(new ServerSideEncryptionCustomerKey(secretKey))
                        .build());
    }

    private static Builder builder() {
        return new Builder();
    }


    // ------------------------------ Builder ------------------------------

    private static class Builder {
        private MinioProp minioProp;

        private Builder(){}

        public Builder minioProp(MinioProp minioProp) {
            this.minioProp = minioProp;
            return this;
        }

        public MinioUtils build() {
            MinioUtils minioUtils = new MinioUtils();
            minioUtils.minioProp = this.minioProp;
            return minioUtils;
        }
    }


}
