
package com.emato.eoss.util.sign;

import javax.crypto.Cipher;
import java.io.ByteArrayOutputStream;
import java.nio.charset.StandardCharsets;
import java.security.*;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;


/**
 * RSA 签名加密
 * RSA 可以签名, 也可以加密
 *
 * @author Scott Chen
 * @date 2017/2/27
 */
public class Rsa {

    public static final String RSA_KEY_ALGORITHM = "RSA";
    public static final String SHA1_RSA_SIGN_ALGORITHMS = "SHA1withRSA";


    private static final String RSA_PUBLIC_KEY = "RSAPublicKey";
    private static final String RSA_PRIVATE_KEY = "RSAPrivateKey";


    /**
     * RSA 密钥长度默认为1024位，密钥长度必须是64的整数倍，范围在512~1024之间
     */
    private static final int KEY_SIZE = 1024;

    /**
     * RSA最大加密明文大小
     */
    private static final int MAX_ENCRYPT_BLOCK = 117;

    /**
     * RSA最大解密密文大小
     */
    private static final int MAX_DECRYPT_BLOCK = 128;


    // ------------------------------ 生成公钥私钥字符串 ------------------------------

    /**
     * 初始化公钥私钥 KeyPair
     *
     * @return
     * @throws Exception
     */
    public static Map<String, Object> initKey() {
        KeyPairGenerator keyPairGen = null;
        try {
            keyPairGen = KeyPairGenerator.getInstance(RSA_KEY_ALGORITHM);
        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
        }

        keyPairGen.initialize(KEY_SIZE);
        KeyPair keyPair = keyPairGen.genKeyPair();

        //获取公私钥对象
        RSAPublicKey publicKey = (RSAPublicKey) keyPair.getPublic();
        RSAPrivateKey privateKey = (RSAPrivateKey) keyPair.getPrivate();

        Map<String, Object> keyMap = new HashMap<>();
        keyMap.put(RSA_PUBLIC_KEY, publicKey);
        keyMap.put(RSA_PRIVATE_KEY, privateKey);

        return keyMap;
    }

    /**
     * 从 KeyPair 中获取公钥字符串
     *
     * @param keyMap
     * @return
     */
    public static String getPublicKeyStr(Map<String, Object> keyMap) {
        Key key = (Key) keyMap.get(RSA_PUBLIC_KEY);
        return Base64.getEncoder().encodeToString(key.getEncoded());
    }

    /**
     * 从 KeyPair 中获取私钥字符串
     *
     * @param keyMap
     * @return
     */
    public static String getPrivateKeyStr(Map<String, Object> keyMap) {
        Key key = (Key) keyMap.get(RSA_PRIVATE_KEY);
        return Base64.getEncoder().encodeToString(key.getEncoded());
    }


    // ------------------------------ 生成公钥私钥 ------------------------------

    /**
     * 生成公钥
     *
     * @param base64PublicKeyStr 密钥字符串（经过base64编码）
     * @return
     * @throws Exception
     */
    public static PublicKey generatePublicKey(String base64PublicKeyStr) throws Exception {
        return generatePublicKey(Base64.getDecoder().decode(base64PublicKeyStr));
    }


    /**
     * 生成公钥
     *
     * @param base64PublicKeyByte 密钥字节码
     * @return
     * @throws Exception
     */
    public static PublicKey generatePublicKey(byte[] base64PublicKeyByte) throws Exception {
        X509EncodedKeySpec keySpec = new X509EncodedKeySpec(base64PublicKeyByte);
        KeyFactory keyFactory = KeyFactory.getInstance(RSA_KEY_ALGORITHM);
        PublicKey publicKey = keyFactory.generatePublic(keySpec);

        return publicKey;
    }


    /**
     * 生成私钥
     *
     * @param base64PrivateKeyStr 密钥字符串（经过base64编码）
     * @return
     * @throws Exception
     */
    public static PrivateKey generatePrivateKey(String base64PrivateKeyStr) throws Exception {
        return generatePrivateKey(Base64.getDecoder().decode(base64PrivateKeyStr));
    }

    /**
     * 生成私钥
     *
     * @param base64PrivateKeyByte 密钥字节码
     * @return
     * @throws Exception
     */
    public static PrivateKey generatePrivateKey(byte[] base64PrivateKeyByte) throws Exception {
        PKCS8EncodedKeySpec keySpec = new PKCS8EncodedKeySpec(base64PrivateKeyByte);
        KeyFactory keyFactory = KeyFactory.getInstance(RSA_KEY_ALGORITHM);
        PrivateKey privateKey = keyFactory.generatePrivate(keySpec);

        return privateKey;
    }



    // ------------------------------ 加签验签 ------------------------------

    /**
     * RSA 签名
     * 私钥签名
     *
     * @param plainStr 待签名字符串
     * @param base64PrivateKeyStr
     * @param inputCharset
     * @return 签名值
     */
    public static String sign(String plainStr, String base64PrivateKeyStr, String inputCharset){
        try
        {
            inputCharset = (inputCharset == null || "".equals(inputCharset)) ? StandardCharsets.UTF_8.toString() : inputCharset;
            return sign(plainStr.getBytes(inputCharset), Base64.getDecoder().decode(base64PrivateKeyStr));
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }


    /**
	 * RSA 签名
	 * 私钥签名
     *
	 * @param plainByte 待签名字节码
	 * @param base64PrivateKeyByte 私钥
	 * @return 签名值
	 */
	public static String sign(byte[] plainByte, byte[] base64PrivateKeyByte) {
        try 
        {
			PrivateKey priKey = generatePrivateKey(base64PrivateKeyByte);

            Signature signature = Signature.getInstance(SHA1_RSA_SIGN_ALGORITHMS);

            signature.initSign(priKey);
            signature.update(plainByte);

            byte[] signed = signature.sign();
            
            return Base64.getEncoder().encodeToString(signed);
        } catch (Exception e) {
        	e.printStackTrace();
        }
        
        return null;
    }


    /**
     * RSA 签名
     * 私钥签名
     *
     * @param plainStr 待签名字符串
     * @param base64SignStr
     * @param base64PublicKeyStr
     * @param inputCharset
     * @return 签名值
     */
    public static boolean verify(String plainStr, String base64SignStr, String base64PublicKeyStr, String inputCharset){
        try
        {
            inputCharset = (inputCharset == null || "".equals(inputCharset)) ? StandardCharsets.UTF_8.toString() : inputCharset;
            return verify(
                    plainStr.getBytes(inputCharset),
                    Base64.getDecoder().decode(base64SignStr),
                    Base64.getDecoder().decode(base64PublicKeyStr)
            );
        } catch (Exception e) {
            e.printStackTrace();
        }

        return false;
    }

	/**
	 * RSA 验签
	 * 公钥验签
	 * @param plainByte 待签名字节码
	 * @param base64SignByte 签名值
	 * @param base64PublicKeyByte 公钥
	 * @return 布尔值
	 */
	public static boolean verify(byte[] plainByte, byte[] base64SignByte, byte[] base64PublicKeyByte) {
		try 
		{
	        PublicKey pubKey = generatePublicKey(base64PublicKeyByte);

			Signature signature = Signature.getInstance(SHA1_RSA_SIGN_ALGORITHMS);
		
			signature.initVerify(pubKey);
			signature.update(plainByte);

			return signature.verify(base64SignByte);
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return false;
	}


	// ------------------------------ 加密解密 ------------------------------

	/**
	 * 加密
	 * 公钥加密
	 * @param plainStr 待加密字符串
	 * @param base64PublicKeyStr 公钥
	 * @param inputCharset
	 * @return 加密字符串
	 * @throws Exception
	 */
	public static String encrypt(String plainStr, String base64PublicKeyStr, String inputCharset) throws Exception {
        inputCharset = (inputCharset == null || "".equals(inputCharset)) ? StandardCharsets.UTF_8.toString() : inputCharset;
	    return encrypt(plainStr.getBytes(inputCharset), Base64.getDecoder().decode(base64PublicKeyStr));
	}

    /**
     * 加密
     * 公钥加密
     * @param plainByte 待加密字节码
     * @param base64PublicKeyByte 公钥
     * @return
     * @throws Exception
     */
    public static String encrypt(byte[] plainByte, byte[] base64PublicKeyByte) throws Exception {
        PublicKey pubKey = generatePublicKey(base64PublicKeyByte);

        Cipher cipher = Cipher.getInstance(RSA_KEY_ALGORITHM);
        cipher.init(Cipher.ENCRYPT_MODE, pubKey);

        ByteArrayOutputStream out = new ByteArrayOutputStream();

        int inputLen = plainByte.length;

        int offSet = 0;
        int i = 0;
        byte[] cache;

        while (inputLen - offSet > 0) {
            if (inputLen - offSet > MAX_ENCRYPT_BLOCK) {
                cache = cipher.doFinal(plainByte, offSet, MAX_ENCRYPT_BLOCK);
            } else {
                cache = cipher.doFinal(plainByte, offSet, inputLen - offSet);
            }
            out.write(cache, 0, cache.length);
            i++;
            offSet = i * MAX_ENCRYPT_BLOCK;
        }

        byte[] encryptByte = out.toByteArray();
        out.close();

        return new String(encryptByte);
    }



    /**
     * 解密
     * 私钥解密
     * @param encryptStr 加密字符串
     * @param base64PrivateKeyStr 私钥
     * @param inputCharset
     * @return 解密后的字符串
     * @throws Exception
     */
    public static String decrypt(String encryptStr , String base64PrivateKeyStr, String inputCharset) throws Exception {
        inputCharset = (inputCharset == null || "".equals(inputCharset)) ? StandardCharsets.UTF_8.toString() : inputCharset;
        return decrypt(encryptStr .getBytes(inputCharset), Base64.getDecoder().decode(base64PrivateKeyStr));
    }

    /**
	 * 解密
	 * 私钥解密
	 * @param encryptByte 加密字节码
	 * @param base64PrivateKeyByte 私钥
	 * @return 解密后的字符串
	 */
	public static String decrypt(byte[] encryptByte, byte[] base64PrivateKeyByte) throws Exception {
        PrivateKey priKey = generatePrivateKey(base64PrivateKeyByte);

        Cipher cipher = Cipher.getInstance(RSA_KEY_ALGORITHM);
        cipher.init(Cipher.DECRYPT_MODE, priKey);

		ByteArrayOutputStream out = new ByteArrayOutputStream();

        //RSA 解密的字节大小最多是128，将需要解密的内容，按128位拆开解密

		int inputLen = encryptByte.length;

		int offSet = 0;
		byte[] cache;
		int i = 0;
		// 对数据分段解密
		while (inputLen - offSet > 0) {
			if (inputLen - offSet > MAX_DECRYPT_BLOCK) {
				cache = cipher.doFinal(encryptByte, offSet, MAX_DECRYPT_BLOCK);
			} else {
				cache = cipher.doFinal(encryptByte, offSet, inputLen - offSet);
			}
			out.write(cache, 0, cache.length);
			i++;
			offSet = i * MAX_DECRYPT_BLOCK;
		}
		byte[] plainByte = out.toByteArray();
		out.close();

        return new String(plainByte);
    }





}
